import copy
from collections import defaultdict
from enum import Enum
import os
import time

DIRECTION = Enum("DIRECTION", ["LEFT", "RIGHT"])
LEFT, RIGHT = DIRECTION.LEFT, DIRECTION.RIGHT

def flip(direction):
    return {LEFT: RIGHT, RIGHT: LEFT}[direction]

def dir2symbol(direction):
    return {LEFT: "/", RIGHT: "\\"}[direction]

COLOR = Enum("COLOR", ["RED", "BLUE"])
RED, BLUE = COLOR.RED, COLOR.BLUE

def color2symbol(color):
    return {RED: "R", BLUE: "B"}[color]

####### cell types that can be activated
class Cell:
    def __init__(self):
        self.location = None
        self.board = None
    def draw(self): raise NotImplementedError
    def activate(self): raise NotImplementedError
    def iter_cardinal_cells(self):
        for delta in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
            location = (self.location[0] + delta[0],
                        self.location[1] + delta[1])
            cell = self.board.cells[location]
            yield cell

class Bit(Cell):
    def __init__(self, initial_direction=LEFT):
        super().__init__()
        self.direction = initial_direction

    def draw(self):
        return dir2symbol(self.direction)

    def activate(self):
        self.board.marble = self.board.marble.move(self.direction)
        self.direction = flip(self.direction)

class Crossover(Cell):
    def draw(self): return "X"
    def activate(self):
        self.board.marble = self.board.marble.move(flip(self.board.marble.corner))

class Ramp(Cell):
    def __init__(self, direction):
        super().__init__()
        self.direction = direction
    def draw(self): return dir2symbol(self.direction)
    def activate(self):
        self.board.marble = self.board.marble.move(self.direction)

class Interceptor(Cell):
    def __init__(self):
        super().__init__()
        self.sequence = []
    def draw(self): return "_"
    def activate(self):
        self.board.sequence.append((self, self.board.marble))
        self.board.marble = None

class GearBit(Bit):
    def activate(self):
        out_dir = self.direction
        new_dir = flip(self.direction)
        for linked_gearbit in self.iter_linked_gearbits():
            assert linked_gearbit.direction == out_dir
            linked_gearbit.direction = new_dir
        self.board.marble = self.board.marble.move(out_dir)

    def iter_linked_gearbits(self):
        worklist = [self]
        seen = set()
        while worklist:
            cell = worklist.pop()
            if cell in seen: continue
            seen.add(cell)
            if isinstance(cell, GearBit): yield cell
            for other in cell.iter_cardinal_cells():
                if isinstance(other, (Gear, GearBit)):
                    worklist.append(other)

class DropTrigger(Cell):
    def __init__(self, marble):
        super().__init__()
        self.marble = marble
        self.sequence = []
    def draw(self): return "T"
    def activate(self):
        self.board.sequence.append((self, self.board.marble))
        self.board.marble = self.marble

####### cell types that shouldn't be activated
class Empty(Cell):
    def draw(self): return " "

class Gear(Cell):
    def draw(self): return "G"

####### board
# locations are (vert, horiz), where larger vert numbers means *lower* on the
# board and larger horiz numbers means more to the right on the board
class Board:
    def __init__(self):
        self.cells = defaultdict(Empty)
        self.marble = None
        self.vert_range = (0, 1)
        self.horiz_range = (0, 1)
        self.sequence = []

    def place_cell(self, location, cell):
        cell.location = location
        cell.board = self
        self.cells[location] = cell

        vert, horiz = location
        self.vert_range = (min(vert, self.vert_range[0]),
                           max(vert + 1, self.vert_range[1]))
        self.horiz_range = (min(horiz, self.horiz_range[0]),
                            max(horiz + 1, self.horiz_range[1]))

    def step(self):
        self = copy.deepcopy(self)
        self.cells[self.marble.location].activate()
        return self

    def draw(self):
        interceptors = []
        for vert in range(*self.vert_range):
            for horiz in range(*self.horiz_range):
                cell = self.cells[(vert, horiz)]
                if isinstance(cell, (Interceptor, DropTrigger)):
                    interceptors.append(cell)
                if self.marble and (vert, horiz) == self.marble.location:
                    if self.marble.color == RED:
                        print(bcolors.FAIL, end="")
                    else:
                        print(bcolors.OKBLUE, end="")
                print(cell.draw(), end="")
                print(bcolors.ENDC, end="")
            print("\n", end="")

        for cell in interceptors:
            print(f"Interceptor at {str(cell.location):>8} got: ", end="")
            for marble_cell, marble in self.sequence:
                marble.start_color()
                if marble_cell is cell:
                    print(color2symbol(marble.color), end=" ")
                else:
                    print(" ", end=" ")
                print(bcolors.ENDC, end="")
            print("\n", end="")

    def interactive(self, auto_stop=False):
        path = [self]
        while True:
            os.system("clear")
            print("State", len(path))
            path[-1].draw()
            cmd = input("Enter to continue, left arrow to go back, stop to stop, 'play x' to autoplay: ")
            if cmd == "\x1b[D":
                path = path[:-1]
            elif cmd == "stop":
                return path[-1]
            elif cmd.startswith("play "):
                secs = float(cmd.split()[1])
                while path[-1].marble:
                    os.system("clear")
                    path.append(path[-1].step())
                    path[-1].draw()
                    time.sleep(secs)
                return path[-1]
            elif path[-1].marble:
                path.append(path[-1].step())
            elif auto_stop:
                return path[-1]

class Marble:
    def __init__(self, location, corner, color):
        self.location = location
        self.corner = corner
        self.color = color

    def move(self, direction):
        vert, horiz = self.location
        if direction == LEFT:
            return Marble((vert + 1, horiz - 1), RIGHT, self.color)
        return Marble((vert + 1, horiz + 1), LEFT, self.color)

    def start_color(self):
        if self.color == RED:
            print(bcolors.FAIL, end="")
        else:
            print(bcolors.OKBLUE, end="")

######### https://stackoverflow.com/questions/287871/how-do-i-print-colored-text-to-the-terminal
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
